# DO NOT EDIT
# This file was generated by idol_py, any changes will be lost when idol_py is rerun again
from ..__idol__ import Struct, Primitive


class SchemaLiteral(Struct):
    bool: bool

    double: float

    int: int

    string: str

    __field_constructors__ = [
        ("bool", "bool", Primitive.of(bool), dict(optional=False),),
        ("double", "double", Primitive.of(float), dict(optional=False),),
        ("int", "int", Primitive.of(int), dict(optional=False),),
        ("string", "string", Primitive.of(str), dict(optional=False),),
    ]
