# DO NOT EDIT
# This file was generated by idol_py, any changes will be lost when idol_py is rerun again
from ..__idol__ import Enum


class SchemaPrimitiveType(Enum):
    INT = "int"
    DOUBLE = "double"
    STRING = "string"
    BOOL = "bool"
    ANY = "any"
