# DO NOT EDIT
# This file was generated by idol_py, any changes will be lost when idol_py is rerun again
from ..__idol__ import Enum


class SchemaStructKind(Enum):
    """
    Scalars are non contained values
    Repeated are homogenous lists.  They can be upgraded or downgraded to and from Scalars.
    A repeated is indicated in a TypeDec by ending with '[]'
    Maps are homogenous 'dictionaries', whose key is always a string, mapping to js objects.
    A map is indicated in a TypeDec by ending with '{}'
    """

    SCALAR = "Scalar"
    REPEATED = "Repeated"
    MAP = "Map"
